package Couchbase::libcouchbase;
use strict;
use warnings;
use File::ShareDir;

our $VERSION = '1.0.4';

__END__

=head1 NAME

Couchbase::libcouchbase - heavy lifting for libcouchbase shared object dependencies


